// VinciaMGInterface.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2014 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Wrappers for the MadGraph F77 matrix elements

#include "VinciaMGInterface.h"

using namespace Pythia8;

namespace Vincia {

  // Wrappers for common blocks, ported from coupl.inc
  
  extern "C" {
    
    extern struct {
      double          g;	
      complex<double> gg[2];	
    } coupl_qcd_;

    extern struct {
      double          hmass, wmass, zmass, amass, tmass, bmass, lmass, cmass;
    } coupl_mass_;
    
    extern struct {
      double          hwidth, wwidth, zwidth, twidth, lwidth, awidth;
    } coupl_width_;
    
    extern struct {
      complex<double> gal[2], gad[2], gau[2], gwf[2], gzn[2], gzl[2], gzd[2], gzu[2];
      double          gw, gwwa, gwwz;      
    } coupl_gauge_;
    
    extern struct {
      complex<double> gwfud[2],  gwfcd[2], gwftd[2],
	gwfus[2],  gwfcs[2], gwfts[2],
	gwfub[2],  gwfcb[2], gwftb[2];
    } coupl_ckm_;
    
    // Uncomment for MadGraph with diagonal CKM:
    /*
    extern struct {
      complex<double> gwfc[2],  gwfs[2], gwfm[2];    
    } coupl_ckm_;
    */

    extern struct {
      complex<double> gwwh, gzzh, gwwhh, gzzhh, ghhh, ghhhh;
    } coupl_scal_;
    
    extern struct {
      complex<double> ghtop[2], ghbot[2], ghtau[2], ghcha[2];
    } coupl_yuk_;

    //Add coupling of higgs to gluons for PS/ME comparison
    extern struct {
      complex<double> gh[2];
    } coupl_heft_;
      
    // Wrappers for MG decay matrix elements

    // Born
    extern void sz2bbx_(int *cMode, double P1[3][4], int HEL[3], double ANS[1]);
    extern void sw2bcx_(int *cMode, double P1[3][4], int HEL[3], double ANS[1]);
    extern void swx2cbx_(int *cMode, double P1[3][4], int HEL[3], double ANS[1]);
    extern void sh2bbx_(int *cMode, double P1[3][4], int HEL[3], double ANS[1]);
    extern void sh2gg_(int *cMode, double P1[3][4], int HEL[3], double ANS[1]);
    extern void shx2bxc_(int *cMode, double P1[3][4], int HEL[3], double ANS[1]);


    // 1st Order
    extern void sz2bgbx_(int *cMode, double P1[4][4], int HEL[3], 
			 double ANS[1]);
    extern void sw2bgcx_(int *cMode, double P1[4][4], int HEL[3], 
			 double ANS[1]); 
    extern void swx2cgbx_(int *cMode, double P1[4][4], int HEL[3], 
			 double ANS[1]); 
    extern void sh2bgbx_(int *cMode, double P1[4][4], int HEL[3],
                           double ANS[1]);
    extern void sh2ggg_(int *cMode, double P1[4][4], int HEL[3], double ANS[1]);
    extern void shx2bxgc_(int *cMode, double P1[4][4], int HEL[3], double ANS[1]);


    // 2nd Order
    extern void sz2bggbx_(int *cMode, double P1[5][4], int HEL[3], 
			  double ANS[1]);
    extern void sz2bcxcbx_(int *cMode, double P1[5][4], int HEL[3], 
			   double ANS[1]);
    extern void sw2bggcx_(int *cMode, double P1[5][4], int HEL[3], 
			  double ANS[1]);
    extern void swx2cggbx_(int *cMode, double P1[5][4], int HEL[3], 
			  double ANS[1]);
    extern void sh2bggbx_(int *cMode, double P1[5][4], int HEL[3],
                            double ANS[1]);
    extern void sh2bcxcbx_(int *cMode, double P1[5][4], int HEL[3],
			   double ANS[1]);
    extern void sh2gggg_(int *cMode, double P1[5][4], int HEL[3],
                         double ANS[1]);
    extern void shx2bxggc_(int *cMode, double P1[5][4], int HEL[3],
                          double ANS[1]);


    // 3rd Order
    extern void sz2bgggbx_(int *cMode, double P1[6][4], int HEL[3], 
			   double ANS[1]);
    extern void sz2bcxcgbx_(int *cMode, double P1[6][4], int HEL[3], 
			    double ANS[1]);
    extern void sz2bgcxcbx_(int *cMode, double P1[6][4], int HEL[3], 
			    double ANS[1]);
    extern void sh2bgggbx_(int *cMode, double P1[6][4], int HEL[3],
                             double ANS[1]);
    extern void sh2bcxcgbx_(int *cMode, double P1[6][4], int HEL[3],
                              double ANS[1]);
    extern void sh2bgcxcbx_(int *cMode, double P1[6][4], int HEL[3],
                              double ANS[1]);
    extern void sh2ggggg_(int *cMode, double P1[6][4], int HEL[3], double ANS[1]);
    
    // 4th Order
    extern void sz2bggggbx_(int *cMode, double P1[7][4], int HEL[3], 
			    double ANS[1]);
    extern void sz2bcxcggbx_(int *cMode, double P1[7][4], int HEL[3], 
			     double ANS[1]);
    extern void sz2bgcxcgbx_(int *cMode, double P1[7][4], int HEL[3], 
			     double ANS[1]);
    extern void sz2bggcxcbx_(int *cMode, double P1[7][4], int HEL[3], 
			     double ANS[1]);
    extern void sz2bcxcuxubx_(int *cMode, double P1[7][4], int HEL[3], 
			      double ANS[1]);
    extern void sz2buxucxcbx_(int *cMode, double P1[7][4], int HEL[3], 
			      double ANS[1]);
    extern void sh2bggggbx_(int *cMode, double P1[7][4], int HEL[3],
                              double ANS[1]);
    extern void sh2bcxcggbx_(int *cMode, double P1[7][4], int HEL[3],
                               double ANS[1]);
    extern void sh2bgcxcgbx_(int *cMode, double P1[7][4], int HEL[3],
                               double ANS[1]);
    extern void sh2bggcxcbx_(int *cMode, double P1[7][4], int HEL[3],
                               double ANS[1]);
    extern void sh2bcxcuxubx_(int *cMode, double P1[7][4], int HEL[3],
                                double ANS[1]);
    extern void sh2buxucxcbx_(int *cMode, double P1[7][4], int HEL[3],
                                double ANS[1]);
    
  }

  //*********

  // Function to swap parameters (eg to use Z->bb matrix element for Z->uu)
  // IMPORTANT NOTE: always remember to swap back after such operations

  void MGInterface::swapZud() {
    complex<double> tmp0 = coupl_gauge_.gzu[0];
    complex<double> tmp1 = coupl_gauge_.gzu[1];
    coupl_gauge_.gzu[0] = coupl_gauge_.gzd[0];
    coupl_gauge_.gzu[1] = coupl_gauge_.gzd[1];
    coupl_gauge_.gzd[0] = tmp0;
    coupl_gauge_.gzd[1] = tmp1;
  }

  //*********

  // Initialize MadGraph
  // Function to initialize the MadGraph coupling common blocks, required
  // by the MG matrix elements. Adapted from couplings.f in the MG 
  // Models/sm/ directory.

  bool MGInterface::init() {
    
    // Set verbose level
    verbose = settingsPtr->mode("Vincia:verbose");

    // If matching not requested and unpolarized shower, just return
    if (settingsPtr->mode("Vincia:matchingLO") <= 0
	&& settingsPtr->flag("Vincia:helicityShower") == 0) return false;
    
    // Verbose output
    if (verbose >= 2) printOut("MGInterface::init","...");

    // Check if pointers initialized
    if (! isInitPtr) {
      printOut("MGInterface::init","Cannot initialize, pointers not set.");
      return false;      
    }
    
    // Innocent until proven guilty
    isInit = true;

    // Verbose output
    if (verbose >= 3) printOut("MGInterface::init","   setting MG masses...");

    // Set masses and widths (initialized through Pythia)
    coupl_mass_.cmass   = particleDataPtr->m0(4);
    coupl_mass_.bmass   = particleDataPtr->m0(5);
    coupl_mass_.tmass   = particleDataPtr->m0(6);
    coupl_mass_.lmass   = particleDataPtr->m0(15);
    coupl_mass_.amass   = particleDataPtr->m0(22);
    coupl_mass_.zmass   = particleDataPtr->m0(23);
    coupl_mass_.wmass   = particleDataPtr->m0(24);
    coupl_mass_.hmass   = particleDataPtr->m0(25);
    coupl_width_.twidth = particleDataPtr->mWidth(6);
    coupl_width_.lwidth = particleDataPtr->mWidth(15);
    coupl_width_.awidth = particleDataPtr->mWidth(22);
    coupl_width_.zwidth = particleDataPtr->mWidth(23);
    coupl_width_.wwidth = particleDataPtr->mWidth(24);
    coupl_width_.hwidth = particleDataPtr->mWidth(25);

    // Useful shorthands
    const double Fourpi = 4.0 * M_PI;
    const double Zero   = 0.0;
    // Here using alpha(mZ) (as in SLHA) rather than alpha(0)
    // Check consistency when doing full matrix elements.
    const double alpha  = settingsPtr->parm("StandardModel:alphaEMmZ");
    // Use thetabar for VFF couplings? (as in Pythia)
    // Check consistency when doing full matrix elements.
    const double sin2w  = couplingsPtr->sin2thetaW();
    const double cw     = sqrt( 1.0 - sin2w );
    const double ee2    = alpha * Fourpi;
    const double sw     = sqrt( sin2w );
    const double ee     = sqrt( ee2 );
    const double ez     = ee/(sw*cw);
    const double ey     = ee*(sw/cw);
    const double sc2    = sin2w*( 1.0 - sin2w );
    const double v      = 2.0*coupl_mass_.wmass*sw/ee;    
    //    const double lambda = pow2(coupl_mass_.hmass) / (2.0 * pow2(v));
    // CKM
    const double Vud    = couplingsPtr->VCKMgen(1,1);
    const double Vus    = couplingsPtr->VCKMgen(1,2);
    const double Vub    = couplingsPtr->VCKMgen(1,3);
    const double Vcd    = couplingsPtr->VCKMgen(2,1);
    const double Vcs    = couplingsPtr->VCKMgen(2,2);
    const double Vcb    = couplingsPtr->VCKMgen(2,3);
    const double Vtd    = couplingsPtr->VCKMgen(3,1);
    const double Vts    = couplingsPtr->VCKMgen(3,2);
    const double Vtb    = couplingsPtr->VCKMgen(3,3);

    // Verbose output
    if (verbose >= 3) printOut("MGInterface::init","   setting MG couplings...");

    // Set QCD g = 1 to keep factored out (=> alphaS = 1/sqrt(4pi))
    coupl_qcd_.g      = 1.0;
    coupl_qcd_.gg[0]  = -coupl_qcd_.g;
    coupl_qcd_.gg[1]  = -coupl_qcd_.g;

    // Set Vector boson couplings
    coupl_gauge_.gw   = ee/sw;
    coupl_gauge_.gwwa = ee;
    coupl_gauge_.gwwz = ee*cw/sw;
    
    // Set gauge & higgs boson couplings
    coupl_scal_.gwwh  = complex<double>( ee2/sin2w*0.5*v, Zero );
    coupl_scal_.gzzh  = complex<double>( ee2/sc2*0.5*v  , Zero );
    coupl_scal_.ghhh  = complex<double>( -pow2(coupl_mass_.hmass)/v*3.0, Zero );
    coupl_scal_.gwwhh = complex<double>( ee2/sin2w*0.5  , Zero );
    coupl_scal_.gzzhh = complex<double>( ee2/sc2*0.5    , Zero );
    coupl_scal_.ghhhh = coupl_scal_.ghhh/v;

    // Set fermion-fermion-vector couplings
    coupl_gauge_.gal[0] = complex<double>(  ee          , Zero );
    coupl_gauge_.gal[1] = complex<double>(  ee          , Zero );
    coupl_gauge_.gau[0] = complex<double>( -ee*2./3.    , Zero );
    coupl_gauge_.gau[1] = complex<double>( -ee*2./3.    , Zero );
    coupl_gauge_.gad[0] = complex<double>(  ee/3.       , Zero );
    coupl_gauge_.gad[1] = complex<double>(  ee/3.       , Zero );

    coupl_gauge_.gzn[0] = complex<double>( -ez*0.5                  , Zero );
    coupl_gauge_.gzn[1] = complex<double>(  Zero                    , Zero );
    coupl_gauge_.gzl[0] = complex<double>( -ez*(-0.5 + sin2w)       , Zero );
    coupl_gauge_.gzl[1] = complex<double>( -ey                      , Zero );
    coupl_gauge_.gzu[0] = complex<double>( -ez*( 0.5 - sin2w*2./3.) , Zero );
    coupl_gauge_.gzu[1] = complex<double>(  ey*2./3.                , Zero );
    coupl_gauge_.gzd[0] = complex<double>( -ez*(-0.5 + sin2w/3.)    , Zero );
    coupl_gauge_.gzd[1] = complex<double>( -ey/3.                   , Zero );

    coupl_gauge_.gwf[0] = complex<double>( -ee/sqrt(2.*sin2w) , Zero );
    coupl_gauge_.gwf[1] = complex<double>(  Zero              , Zero );

    // CKM
    for (int iLR = 0; iLR <= 1; ++iLR) {
      coupl_ckm_.gwfud[iLR] = Vud * coupl_gauge_.gwf[iLR]; 
      coupl_ckm_.gwfus[iLR] = Vus * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwfub[iLR] = Vub * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwfcd[iLR] = Vcd * coupl_gauge_.gwf[iLR]; 
      coupl_ckm_.gwfcs[iLR] = Vcs * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwfcb[iLR] = Vcb * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwftd[iLR] = Vtd * coupl_gauge_.gwf[iLR]; 
      coupl_ckm_.gwfts[iLR] = Vts * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwftb[iLR] = Vtb * coupl_gauge_.gwf[iLR];
   }

    // Should also fill in Yukawa couplings, but more tricky due
    // to running, so take up again later.
    //Including coupling of bottom to higgs
    coupl_yuk_.ghbot[0] = complex<double>( 1.0       , Zero );
    coupl_yuk_.ghbot[1] = complex<double>( 1.0       , Zero );
    
    //Include coupling of Higgs to gluons
    //Set arbitrarily for ME/PS comparison
    //The small value is set to reflect the loop suppression,
    //etc., of the higgs coupling and decreases the size of the ME.
    coupl_heft_.gh[0] = complex<double>( 0.001      , Zero );
    coupl_heft_.gh[1] = complex<double>( 0.001       , Zero );

    // Verbose output
    if (verbose >= 3) printOut("MGInterface::init","   setting Color Depth...");

    // Set color depth
    colorDepth = (settingsPtr->flag("Vincia:matchingFullColor")) ? 1 : 0;

    return true;
  }
  
  //*********

  // Wrapper to set values in the MG common blocks

  void MGInterface::setParm(string txt, complex<double> val) {

    // Check if initialized
    if (! isInit) {
      printOut("MGInterface::setParm","MGInterface not initialized.");
    }
    // Convert to all lower-case
    for (unsigned int i=0;i<txt.length();++i) txt[i]=tolower(txt[i]);
    if (txt == "zmass" || txt == "mz") coupl_mass_.zmass   = val.real();
    else if (txt == "wmass" || txt == "mw") coupl_mass_.wmass   = val.real();
    else if (txt == "bmass" || txt == "mb") coupl_mass_.bmass   = val.real();
    else if (txt == "hmass" || txt == "mh") coupl_mass_.hmass   = val.real();
    // A (photon) couplings
    else if (txt == "gau0") coupl_gauge_.gau[0]=val;
    else if (txt == "gau1") coupl_gauge_.gau[1]=val;
    else if (txt == "gad0") coupl_gauge_.gad[0]=val;
    else if (txt == "gad1") coupl_gauge_.gad[1]=val;
    else if (txt == "gal0") coupl_gauge_.gal[0]=val;
    else if (txt == "gal1") coupl_gauge_.gal[1]=val;
    // Z couplings
    else if (txt == "gzu0") coupl_gauge_.gzu[0]=val;
    else if (txt == "gzu1") coupl_gauge_.gzu[1]=val;
    else if (txt == "gzd0") coupl_gauge_.gzd[0]=val;
    else if (txt == "gzd1") coupl_gauge_.gzd[1]=val;
    else if (txt == "gzn0") coupl_gauge_.gzn[0]=val;
    else if (txt == "gzn1") coupl_gauge_.gzn[1]=val;
    else if (txt == "gzl0") coupl_gauge_.gzl[0]=val;
    else if (txt == "gzl1") coupl_gauge_.gzl[1]=val;
    // W couplings
    else if (txt.find("gwf") != string::npos) {
      if (txt == "gwf0") coupl_gauge_.gwf[0]=val;
      else if (txt == "gwf1") coupl_gauge_.gwf[1]=val;
      // CKM
      else if (txt == "gwfud0") coupl_ckm_.gwfud[0]=val;
      else if (txt == "gwfud1") coupl_ckm_.gwfud[1]=val;
      else if (txt == "gwfcd0") coupl_ckm_.gwfcd[0]=val;
      else if (txt == "gwfcd1") coupl_ckm_.gwfcd[1]=val;
      else if (txt == "gwftd0") coupl_ckm_.gwftd[0]=val;
      else if (txt == "gwftd1") coupl_ckm_.gwftd[1]=val;
      else if (txt == "gwfus0") coupl_ckm_.gwfus[0]=val;
      else if (txt == "gwfus1") coupl_ckm_.gwfus[1]=val;
      else if (txt == "gwfcs0") coupl_ckm_.gwfcs[0]=val;
      else if (txt == "gwfcs1") coupl_ckm_.gwfcs[1]=val;
      else if (txt == "gwfts0") coupl_ckm_.gwfts[0]=val;
      else if (txt == "gwfts1") coupl_ckm_.gwfts[1]=val;
      else if (txt == "gwfub0") coupl_ckm_.gwfub[0]=val;
      else if (txt == "gwfub1") coupl_ckm_.gwfub[1]=val;
      else if (txt == "gwfcb0") coupl_ckm_.gwfcb[0]=val;
      else if (txt == "gwfcb1") coupl_ckm_.gwfcb[1]=val;
      else if (txt == "gwftb0") coupl_ckm_.gwftb[0]=val;
      else if (txt == "gwftb1") coupl_ckm_.gwftb[1]=val;
    }
    // H couplings
    else if (txt == "ghbot0") coupl_yuk_.ghbot[0] = val;
    else if (txt == "ghbot1") coupl_yuk_.ghbot[1] = val;
    else if (txt == "gh0") coupl_heft_.gh[0] = val;
    else if (txt == "gh1") coupl_heft_.gh[1] = val;
    // Else issue error message: unrecognized coupling parameter
    else {
      printOut("MGInterface::setParm","Unknown parameter "+txt);
    }

  }

  void MGInterface::setParm(string txt, double val) {

    complex<double> valC(val,0.0);
    setParm(txt, valC);

  }

  //*********

  // Wrapper to return unpolarized Madgraph Matrix Elements (squared)
  double MGInterface::ME2(int nIn, vector<int> id, vector<Vec4> p) {

    vector<double> pol;
    for (unsigned int i=0; i<p.size(); ++i) pol.push_back(9.);
    return ME2(nIn, id, p, pol);

  }

  // Wrapper to return polarized MadGraph Matrix Elements (squared)
  // Input:
  //    nIn   = number of particles that are incoming (1 or 2)
  //    id    = vector of particle IDs. The incoming particle(s) MUST be the 
  //            first nIn entries of this vector. The outgoing particles
  //            MUST be arranged in color order thereafter. 
  //    p     = vector of particle momenta. The incoming particle(s) MUST be 
  //            the first nIn entries of this vector. The outgoing particles
  //            MUST be arranged in color order thereafter. 
  //	pol   = vector of particle polarizations.  pol = 9 is unpolarized.
  //	        The incoming particle(s) MUST be the first nIn entries of 
  //            this vector. The outgoing particles MUST be arranged in 
  //	        color order thereafter.

  // Return:   value of Matrix Element squared for this phase-space point

  double MGInterface::ME2(int nIn, vector<int> id, vector<Vec4> p, 
			  vector<int> pol) {

  
    // Check if initialized
    if (! isInit) {
      printOut("MGInterface::setParm","MGInterface not initialized.");
    }
    
    // Initialize ME weight with dummy value.
    const  int nMom  = p.size();
    const  int nOut  = nMom-nIn;
    double MGME2 = -1.;

    // Take out polarization averages
    int nPolIn = 1;
    
    // Only DECAYS implemented so far
    if (nIn != 1 || int(id.size()) != nMom || nOut <= 1) return MGME2;

    //------------------------------------------------------------------------

    int id0 = abs(id[0]);
    // If not known, return.
    if (id0 == 0) return -1;

    // Check if using massive matrix elements
    bool massless=true;
    double sCM = p[0].m2Calc(); 
    // Fill vector of desired masses (zero ones that are offshell only by
    // an amount consistent with numerical precision)
    vector<double> m;    
    m.push_back(sqrt(sCM));
    for (unsigned int i=1;i<id.size(); ++i) {
      double m2 = p[i].m2Calc();      
      if (abs(m2/sCM) > TINY) {
	massless = false;
	m.push_back(sqrt(abs(m2)));
      } else {
	m.push_back(0.0);
      }
    }
    
    // Verbose output
    if (verbose >= 8) 
      printOut("VinciaMGInterface::ME2","massless approximation = "
	       +bool2str(massless));

    // Check Lorentz invariance 
    // (change to (true) to always boost back to CM) 
    if (false) {
      Vec4 vCM;
      if (nIn == 1) vCM = p[0];
      else vCM = p[0]+p[1];
      double eCM = vCM.e();
      if (eCM <= 0.0) {
	printOut("VinciaMGInterface::ME2",
		 "Error: failed to evaluate ME, eCM <= 0.0 ");
	return -1;
      }
      if (abs(vCM.px()/eCM) + abs(vCM.py()/eCM) + abs(vCM.pz()/eCM) > TINY) {
	cout<<"We are in a boosted frame. Boosting back"<<endl;
	for (unsigned int i=0; i<p.size(); ++i) p[i].bstback(vCM);
      }
    }
    
    //------------------------------------------------------------------------
    // Color-singlet spin-1 particles (from Z and W decays)
    if (id0 == 23 || id0 == 24) {

      // Default: sum over all incoming and outgoing polarizations
      nPolIn = 3;
      int HEL[3];
      for (int i=0; i<nPolIn; ++i) HEL[i]=-1; 

      // Ensure ME evaluated on shell
      const double mSys = p[0].mCalc();
      coupl_mass_.zmass = mSys;
      coupl_mass_.wmass = mSys;
      
      // V->2
      if (nOut == 2) {

	// Convert momenta to MadGraph format
	double PNEW[3][4];
	for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);
	
	// HELCOMB = 1 + (pol[0] + 1)*4 + (pol[2] + 1)/2 + (pol[1] + 1)
	// 0) Default assumption : fully unpolarized Z/W decay
	// 1) Sum over W/Z polarizations, for specific final-state polarizations
	if (pol[0] == 9 && pol[1] != 9 && pol[2] != 9) {
	  HEL[0] = (int)(1 + (pol[2] + 1)/2 + (pol[1] + 1));
	  HEL[1] = (int)(5 + (pol[2] + 1)/2 + (pol[1] + 1));
	  HEL[2] = (int)(9 + (pol[2] + 1)/2 + (pol[1] + 1));
	}
	// 2) Fully polarized Z/W decay (nPolIn = 1 in normalization)
	else if (pol[0] != 9 && pol[1] != 9 && pol[2] != 9) {
	  nPolIn = 1;
	  HEL[0] = (int)(1 + (pol[0] + 1)*4 + (pol[2] + 1)/2 + (pol[1] + 1));
	  HEL[1] = -1;
	  HEL[2] = -1;
	}

	// Evaluate Matrix Element
        if (id0 == 23) {
          if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
          // Massive (equal masses, from Z->bbbar)
          coupl_mass_.bmass = p[1].mCalc();
	  if (abs(id[1]) % 2 == 0) swapZud();
	  sz2bbx_(&colorDepth,PNEW,HEL,&MGME2);
	  if (abs(id[1]) % 2 == 0) swapZud();
	  // For tests:
	  // cout<<"E   = "<<p[0].e()<<" "<<p[1].e()<<" "<<p[2].e()<<endl;
	  // cout<<"HEL = "<<HEL[0]<<" "<<HEL[1]<<" "<<HEL[2]<<" : MGME2 = "
	  //    <<MGME2<<endl;
        }
        else if (id0 == 24 && abs(id[1]) % 2 == 1) {
          if (verbose >= 7) printOut("MGinterface:ME2","using W decay ME");
	  // Massive (unequal masses, from W->bcx)
	  coupl_mass_.bmass = m[1];
	  coupl_mass_.cmass = m[2];
	  sw2bcx_(&colorDepth,PNEW,HEL,&MGME2);
	  // For tests:
	  // cout<<"E   = "<<p[0].e()<<" "<<p[1].e()<<" "<<p[2].e()<<endl;
	  //cout<<"HEL = "<<HEL[0]<<" "<<HEL[1]<<" "<<HEL[2]<<" : MGME2 = "
	  //    <<MGME2<<endl;
	} 
	else if (id0 == 24 && abs(id[1]) % 2 == 0){
          if (verbose >= 7) printOut("MGinterface:ME2","using W decay ME");
	  // Massive (unequal masses, from W+->cbx)
	  coupl_mass_.cmass = m[1];
	  coupl_mass_.bmass = m[2];
	  swx2cbx_(&colorDepth,PNEW,HEL,&MGME2);
	}
	if (verbose >= 8) {
	  printOut("VinciaMGInterface::ME2","Getting M2 with colDepth = "
		   +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	}

      }

      // V->3
      else if (nOut == 3) {
	// Convert momenta to MadGraph format
	double PNEW[4][4];
	for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);
	
	// HELCOMB = 1 + 8*(pol[0]+1) + 2*(pol[1]+1) + (pol[2]+1) + (pol[3]+1)/2 
 	// 0) Default assumption : fully unpolarized Z/W decay
	// 1) Sum over W/Z polarizations, for specific final-state polarizations
	if (pol[0] == 9 && pol[1] != 9 && pol[2] != 9 && pol[3] != 9) {
	  HEL[0] = (int)(1 + (pol[3] + 1)/2 + (pol[2] + 1) + 2*(pol[1] + 1));
	  HEL[1] = (int)(9 + (pol[3] + 1)/2 + (pol[2] + 1) + 2*(pol[1] + 1));
	  HEL[2] = (int)(17 + (pol[3] + 1)/2 + (pol[2] + 1) + 2*(pol[1] + 1));
	}
	// 2) Fully polarized Z/W decay (nPolIn = 1 for normalization)
	else if (pol[0] != 9 && pol[1] != 9 && pol[2] != 9 && pol[3] != 9) {
	  nPolIn = 1;
	  HEL[0] = (int)(1 + (pol[0] + 1)*8 + (pol[3] + 1)/2 + (pol[2] + 1) 
			 + 2*(pol[1] + 1));
	  HEL[1] = -1;
	  HEL[2] = -1;
	}

	// Evaluate Matrix Element
	if (id0 == 23) {
          if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
	  // Massive (equal masses, from Z->bgbbar)
	  coupl_mass_.bmass = m[1];
	  if (abs(id[1]) % 2 == 0) swapZud();
	  sz2bgbx_(&colorDepth,PNEW,HEL,&MGME2);
	  if (abs(id[1]) % 2 == 0) swapZud(); 
       	} 
	else if (id0 == 24 && abs(id[1]) % 2 == 1) {
          if (verbose >= 7) printOut("MGinterface:ME2","using W decay ME");
	  // Massive (unequal masses, from W->cgbbar)
	  coupl_mass_.bmass = m[1];
	  coupl_mass_.cmass = m[3];
	  sw2bgcx_(&colorDepth,PNEW,HEL,&MGME2);
	} 
        else if (id0 == 24 && abs(id[1]) % 2 == 0) {	
          if (verbose >= 7) printOut("MGinterface:ME2","using W decay ME");
	  // Massive (unequal masses, from W->cgbbar)
	  coupl_mass_.cmass = m[1];
	  coupl_mass_.bmass = m[3];
	  swx2cgbx_(&colorDepth,PNEW,HEL,&MGME2);
	}

	if (verbose >= 8) {
	  printOut("VinciaMGInterface::ME2","Getting M3 with colDepth = "
		   +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	}

      }

      // Higher than 3, need to distinguish how many g's and how many q's
      else {
	int cSum = 0;
        // colType according to Pythia documentation:
        // the colour type, with 0 uncoloured, 1 triplet, 
        // -1 antitriplet and 2 octet
	for (int i=1; i<nMom; ++i) cSum += particleDataPtr->colType(id[i]);

	// Z->4
	if (nOut == 4) {

	  // Convert momenta to MadGraph format
	  double PNEW[5][4];
	  for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);
	  
	  // HELCOMB = 1 + 16*(pol[0]+1) + 4*(pol[1]+1) + ... + (pol[4] + 1)/2
	  // 0) Default assumption : fully unpolarized Z/W decay
	  // 1) Sum over W/Z polarizations, for specific final-state polarizations
	  if (pol[0] == 9 && pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
	      && pol[4] != 9) {
	    HEL[0] = (int)(1 + (pol[4] + 1)/2 + (pol[3] + 1) + 2*(pol[2] + 1) 
			    + 4*(pol[1] + 1));
	    HEL[1] = (int)(17 + (pol[4] + 1)/2 + (pol[3] + 1) + 2*(pol[2] + 1) 
			   + 4*(pol[1] + 1));
	    HEL[2] = (int)(33 + (pol[4] + 1)/2 + (pol[3] + 1) + 2*(pol[2] + 1) 
			   + 4*(pol[1] + 1));
	  }
	  //2) Fully polarized Z/W decay (-> nPolIn = 1 for normalization)
	  else if (pol[0] != 9 && pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
		   && pol[4] != 9) {
	    nPolIn = 1;
	    HEL[0] = (int)(1 + (pol[0] + 1)*16 + (pol[4] + 1)/2 + (pol[3] + 1) 
			   + 2*(pol[2] + 1) + 4*(pol[1] + 1));
	    HEL[1] = -1;
	    HEL[2] = -1;
	  }

	  // V->qggqbar
	  if (cSum == 4) { 
	    
	    // Evaluate Matrix Element	    
	    if (id0 == 23) {
              if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
	      coupl_mass_.bmass = m[1];
	      if (abs(id[1]) % 2 == 0) swapZud();
	      sz2bggbx_(&colorDepth,PNEW,HEL,&MGME2);
	      if (abs(id[1]) % 2 == 0) swapZud();
	    } 
	    else if (id0 == 24 && abs(id[1]) % 2 == 1 ) {
              if (verbose >= 7) printOut("MGinterface:ME2","using W decay ME");
	      coupl_mass_.bmass = m[1];
	      coupl_mass_.cmass = m[4];
	      sw2bggcx_(&colorDepth,PNEW,HEL,&MGME2);
	    } 
	    else if (id0 == 24 && abs(id[1]) % 2 == 0 ) {
              if (verbose >= 7) printOut("MGinterface:ME2","using W decay ME");
	      coupl_mass_.bmass = m[4];
	      coupl_mass_.cmass = m[1];
	      swx2cggbx_(&colorDepth,PNEW,HEL,&MGME2);
	    }

	    if (verbose >= 8) {
	      printOut("VinciaMGInterface::ME2","Getting M4 with colDepth = "
		       +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	    }
	  } 

	  // Z->qq'barq'qbar
	  else if (cSum == 0){

	    // Evaluate Matrix Element: Z decay or polarized W decay
	    if (id0 == 23) {
              if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
	      // the matrix element has no up-type z coupling
	      coupl_mass_.bmass = m[1];
	      coupl_mass_.cmass = m[2];
	      if (abs(id[1]) % 2 == 0) swapZud();
	      sz2bcxcbx_(&colorDepth,PNEW,HEL,&MGME2);
	      if (abs(id[1]) % 2 == 0) swapZud();
	    }

	    if (verbose >= 8) {
	      printOut("VinciaMGInterface::ME2","Getting M4 with colDepth = "
		       +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	    }
	  }
	}

	// Z->5 or polarized W->5 
	else if (nOut == 5) {

	  // Convert momenta to MadGraph format
	  double PNEW[6][4];
	  for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);


	  // HELCOMB = 1 + (pol[0] + 1)*32 + (pol[1] + 1)*8 + ... + (pol[5] + 1)/2 
	  // 0) Default assumption : fully unpolarized Z/W decay
	  // 1) Sum over W/Z polarizations, for specific final-state polarizations
	  if (pol[0] == 9 && pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
	      && pol[4] != 9 && pol[5] != 9) {
	  HEL[0] = (int)(1 + (pol[5] + 1)/2 + (pol[4] + 1) + 2*(pol[3] + 1) 
			 + 4*(pol[2] + 1) + 8*(pol[1] + 1));
	  HEL[1] = (int)(33 + (pol[5] + 1)/2 + (pol[4] + 1) + 2*(pol[3] + 1) 
			 + 4*(pol[2] + 1) + 8*(pol[1] + 1));
	  HEL[2] = (int)(65 + (pol[5] + 1)/2 + (pol[4] + 1) + 2*(pol[3] + 1) 
			 + 4*(pol[2] + 1) + 8*(pol[1] + 1));
	  }
	  // 2) Fully polarized Z/W decay (-> nPolIn = 1 for normalization)
	  else if (pol[0] != 9 && pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
		   && pol[4] != 9 && pol[5] != 9) {
	    nPolIn = 1;
	    HEL[0] = (int)(1 + (pol[0] + 1)*32 + (pol[5] + 1)/2 + (pol[4] + 1) 
			   + 2*(pol[3] + 1) + 4*(pol[2] + 1) + 8*(pol[1] + 1));
	    HEL[1] = -1;
	    HEL[2] = -1;
	  }
	  
	  // Z->qgggqbar
	  if (cSum == 6 && id0 == 23) {
            if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
            // MR 7.3.2011: set b mass to mass of quark 1
            coupl_mass_.bmass = m[1];
	    if (abs(id[1]) % 2 == 0) swapZud();
	    sz2bgggbx_(&colorDepth,PNEW,HEL,&MGME2);
	    if (abs(id[1]) % 2 == 0) swapZud();
	  }

	  // Z->qgqbar'q'qbar and Z->qqbar'q'gqbar
	  else if (cSum == 2 && id0 == 23) {
            if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
            // Z->qgqbar'q'qbar
	    if (id[2] == 21) {
              coupl_mass_.bmass = m[1];
              coupl_mass_.cmass = m[3];
	      if (abs(id[1]) % 2 == 0) swapZud();
	      sz2bgcxcbx_(&colorDepth,PNEW,HEL,&MGME2);
	      if (abs(id[1]) % 2 == 0) swapZud();
	    }
	    // Z->qqbar'q'gqbar
	    else if (id[4] == 21) {
              coupl_mass_.bmass = m[1];
              coupl_mass_.cmass = m[2];
	      if (abs(id[1]) % 2 == 0) swapZud();
	      sz2bcxcgbx_(&colorDepth,PNEW,HEL,&MGME2);
	      if (abs(id[1]) % 2 == 0) swapZud();
	    }
          }
          if (verbose >= 8) {
	      printOut("VinciaMGInterface::ME2","Getting M5 with colDepth = "
		       +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
          }
	} // end nOut == 5

	// Z->6 or polarized W->6
	else if (nOut == 6) {

	  // Convert momenta to MadGraph format
	  double PNEW[7][4];
	  for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);

	  // HELCOMB = 1 + (pol[0] + 1)*64 + (pol[1] + 1)*16 + ... + (pol[6] + 1)/2
	  // 0) Default assumption : fully unpolarized Z/W decay
	  // 1) Sum over W/Z polarizations, for specific final-state polarizations
	  if (pol[0] == 9 && pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
	      && pol[4] != 9 && pol[5] != 9 && pol[6] != 9) {
	  HEL[0] = (int)(1 + (pol[6] + 1)/2 + (pol[5] + 1) + 2*(pol[4] + 1) 
			 + 4*(pol[3] + 1) + 8*(pol[2] + 1) + 16*(pol[1] + 1));
	  HEL[1] = (int)(65 + (pol[6] + 1)/2 + (pol[5] + 1) + 2*(pol[4] + 1) 
			 + 4*(pol[3] + 1) + 8*(pol[2] + 1) + 16*(pol[1] + 1));
	  HEL[2] = (int)(129 + (pol[6] + 1)/2 + (pol[5] + 1) + 2*(pol[4] + 1) 
			 + 4*(pol[3] + 1) + 8*(pol[2] + 1) + 16*(pol[1] + 1));
	  }
	  //fully polarized Z/W decay
	  else if (pol[0] != 9 && pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
		   && pol[4] != 9 && pol[5] != 9 && pol[6] != 9) {
	    HEL[0] = (int)(1 + (pol[0] + 1)*64 + (pol[6] + 1)/2 + (pol[5] + 1) 
			   + 2*(pol[4] + 1) + 4*(pol[3] + 1) + 8*(pol[2] + 1) 
			   + 16*(pol[1] + 1));
	    HEL[1] = -1;
	    HEL[2] = -1;
	  }

	  // Z->qggggqbar
	  if (cSum == 8 && id0 == 23) {
            if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
            // set quark mass
	    coupl_mass_.bmass = m[1];
	    if (abs(id[1]) % 2 == 0) swapZud();
            sz2bggggbx_(&colorDepth,PNEW,HEL,&MGME2);
	    if (abs(id[1]) % 2 == 0) swapZud();
	  }
          // Z->q qbar' q' g g qbar
          else if (cSum == 4 && (id[2]+id[3]==0) && id0 == 23) {
            if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
            // the matrix element has no Z-u-type coupling
            // set b-mass to that of primary quark, c to secondary
            coupl_mass_.bmass = m[1];
            coupl_mass_.cmass = m[2];
	    if (abs(id[1]) % 2 == 0) swapZud();
            sz2bcxcggbx_(&colorDepth,PNEW,HEL,&MGME2);          
	    if (abs(id[1]) % 2 == 0) swapZud();
          }
          // Z->q g qbar' q' g qbar
          else if (cSum == 4 && (id[3]+id[4]==0) && id0 == 23) {
            if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
            // the matrix element has no Z-u-type coupling
            // set b-mass to that of primary quark, c to secondary
            coupl_mass_.bmass = m[1];
            coupl_mass_.cmass = m[3];
	    if (abs(id[1]) % 2 == 0) swapZud();
            sz2bgcxcgbx_(&colorDepth,PNEW,HEL,&MGME2);
	    if (abs(id[1]) % 2 == 0) swapZud();
          }
          // Z->q g g qbar' q' qbar
          else if (cSum == 4 && (id[4]+id[5]==0) && id0 == 23) {
            if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
            // the matrix element has no Z-u-type coupling
            // set b-mass to that of primary quark, c to secondary
            coupl_mass_.bmass = m[1];
            coupl_mass_.cmass = m[4];
	    if (abs(id[1]) % 2 == 0) swapZud();
	    sz2bggcxcbx_(&colorDepth,PNEW,HEL,&MGME2);
	    if (abs(id[1]) % 2 == 0) swapZud();
          }
          // Z->q q' qbar' q'' qbar'' qbar
          // the MG interface does not allow at present to make a 
          // light quark massive, therefore one of the secondary quarks 
          // involved has to be massless
          else if (cSum == 0 && (id[2]+id[3]==0) && (id[4]+id[5]==0) && id0 == 23) {
            if (verbose >= 7) printOut("MGinterface:ME2","using Z decay ME");
            // the matrix element has no Z-u-type coupling
            if (m[4] == 0.0) {
              coupl_mass_.bmass = m[1];
              coupl_mass_.cmass = m[2];
	      if (abs(id[1]) % 2 == 0) swapZud();
              sz2bcxcuxubx_(&colorDepth,PNEW,HEL,&MGME2);
	      if (abs(id[1]) % 2 == 0) swapZud();
            }
            else if (m[2] == 0) {
              coupl_mass_.bmass = m[1];
              coupl_mass_.cmass = m[4];
	      if (abs(id[1]) % 2 == 0) swapZud();
              sz2buxucxcbx_(&colorDepth,PNEW,HEL,&MGME2);
	      if (abs(id[1]) % 2 == 0) swapZud();
            }
            // case of both secondary quarks massive not implemented
	    else {
	      printOut("VinciaMGInterface::ME2","MEs for double massive-quark pair creation not yet implemented");
	    }
          }
	  
	  if (verbose >= 8) {
	    printOut("VinciaMGInterface::ME2","Getting M6 with colDepth = "
		     +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	  }
	}
	
      }
      
    }

    //------------------------------------------------------------------------
    // Color-singlet spin-0 particles (from higgs decays)
    else if (id0 == 25) {
      if (verbose >= 7) printOut("MGinterface:ME2","using H decay ME");

      // Default: sum over all incoming and outgoing polarizations
      nPolIn = 1;
      int HEL[3];
      for (int i=0; i<nPolIn; ++i) HEL[i]=-1; 

      // Ensure ME evaluated on shell
      coupl_mass_.hmass = m[0];
      
      //Evaluate color structure
      int cSum = 0;
      // colType according to Pythia documentation:
      // the colour type, with 0 uncoloured, 1 triplet,
      // -1 antitriplet and 2 octet
      for (int i=1; i<nMom; ++i) cSum += particleDataPtr->colType(id[i]);
        
      // h->2
      if (nOut == 2) {

        // Convert momenta to MadGraph format
        double PNEW[3][4];
        for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);

        // HELCOMB = 1 + (pol[2] + 1)/2 + (pol[1] + 1)
        // 0) Default assumption : fully unpolarized h decay
        // 1) Fully polarized h decay
        if (pol[1] != 9 && pol[2] != 9) {
          HEL[0] = (int)(1 + (pol[2] + 1)/2 + (pol[1] + 1));
          HEL[1] = -1;
          HEL[2] = -1;
        }

        if (abs(id[1]) == abs(id[2])) {
          //h > gg
          if (cSum == 4) {
            //Evaluate matrix element
            sh2gg_(&colorDepth,PNEW,HEL,&MGME2);
          }
          
          //h > qqbar
          else {
            //Evaluate matrix element
            coupl_mass_.bmass = m[1];
            sh2bbx_(&colorDepth,PNEW,HEL,&MGME2);
          }
        }
        else {
          //h+ > b~ c
          coupl_mass_.bmass = m[1];
          coupl_mass_.cmass = m[2];
          shx2bxc_(&colorDepth,PNEW,HEL,&MGME2);
        }
            
        if (verbose >= 8) {
            printOut("VinciaMGInterface::ME2","Getting M2 with colDepth = "
                    +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
        }
      }
      
      // h->3
      else if (nOut == 3) {

        // Convert momenta to MadGraph format
        double PNEW[4][4];
        for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);

        // HELCOMB = 1 + 2*(pol[1]+1) + (pol[2]+1) + (pol[3]+1)/2 
        // 0) Default assumption : fully unpolarized h decay
        // 1) Fully polarized h decay
        if (pol[1] != 9 && pol[2] != 9 && pol[3] != 9) {
          HEL[0] = (int)(1 + (pol[3] + 1)/2 + (pol[2] + 1) 
                         + 2*(pol[1] + 1));
          HEL[1] = -1;
          HEL[2] = -1;
        }
          
        if (abs(id[1]) == abs(id[3])) {
          //h > ggg
          if (cSum == 6) {
              //Evaluate matrix element
              sh2ggg_(&colorDepth,PNEW,HEL,&MGME2);
          }
          
          //h > qgqbar
          else {
            //Evaluate matrix element
            coupl_mass_.bmass = m[1];
            sh2bgbx_(&colorDepth,PNEW,HEL,&MGME2);
          }
        }
        else {
          //h+ > b~ g c
          coupl_mass_.bmass = m[1];
          coupl_mass_.cmass = m[3];
          shx2bxgc_(&colorDepth,PNEW,HEL,&MGME2);
        }

        if (verbose >= 8) {
            printOut("VinciaMGInterface::ME2","Getting M3 with colDepth = "
                    +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
        }
      }

      // h->4
      if (nOut == 4) {

        // Convert momenta to MadGraph format
        double PNEW[5][4];
        for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);
        
        // HELCOMB = 1 + 4*(pol[1]+1) + ... + (pol[4] + 1)/2
        // 0) Default assumption : fully unpolarized h decay
        //1) Fully polarized h decay
        if (pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
                 && pol[4] != 9) {
          HEL[0] = (int)(1 + (pol[4] + 1)/2 + (pol[3] + 1) 
                         + 2*(pol[2] + 1) + 4*(pol[1] + 1));
          HEL[1] = -1;
          HEL[2] = -1;
        }
            
        if (abs(id[1]) == abs(id[4])) {
          // h->gggg
	  if (cSum == 8) {
	    // Evaluate Matrix Element
            sh2gggg_(&colorDepth,PNEW,HEL,&MGME2);
	  }
          
          // h->qggqbar
	  else if (cSum == 4) {
	    // Evaluate Matrix Element
	    coupl_mass_.bmass = m[1];
            sh2bggbx_(&colorDepth,PNEW,HEL,&MGME2);
	  }
	  
	  // h->qq'barq'qbar
	  else {
	    // the matrix element has no h-u-type coupling
	    // Evaluate Matrix Element
	    coupl_mass_.bmass = m[1];
	    coupl_mass_.cmass = m[2];
            sh2bcxcbx_(&colorDepth,PNEW,HEL,&MGME2);
	  }
        }
        else {
          //h+ > b~ g g c
          coupl_mass_.bmass = m[1];
          coupl_mass_.cmass = m[4];
          shx2bxggc_(&colorDepth,PNEW,HEL,&MGME2);
        }
	
        if (verbose >= 8) {
	  printOut("VinciaMGInterface::ME2","Getting M4 with colDepth = "
                    +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
        }
      }
      
      // h->5 
      else if (nOut == 5) {
	
        // Convert momenta to MadGraph format
        double PNEW[6][4];
        for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);

        // HELCOMB = 1 + (pol[1] + 1)*8 + ... + (pol[5] + 1)/2 
        // 0) Default assumption : fully unpolarized h decay
        // 1) Fully polarized h decay
        if (pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
                 && pol[4] != 9 && pol[5] != 9) {
          HEL[0] = (int)(1 + (pol[5] + 1)/2 + (pol[4] + 1) 
                         + 2*(pol[3] + 1) + 4*(pol[2] + 1) + 8*(pol[1] + 1));
          HEL[1] = -1;
          HEL[2] = -1;
        }
          
        //h>ggggg
        if (cSum == 10) {
          //Evaluate matrix element
          sh2ggggg_(&colorDepth,PNEW,HEL,&MGME2);
        }
            
        //h>qgggqbar
        else if (cSum == 6) {
          //Evaluate matrix element
          coupl_mass_.bmass = m[1];
          if (id[1] != 21 && id[5] != 21) 
	    sh2bgggbx_(&colorDepth,PNEW,HEL,&MGME2);
        }
	
        // h->qgqbar'q'qbar and h->qqbar'q'gqbar
        else if (cSum == 2){
	  // h->qgqbar'q'qbar
	  if (id[2] == 21) {
	    coupl_mass_.bmass = m[1];
	    coupl_mass_.cmass = m[3];
	    sh2bgcxcbx_(&colorDepth,PNEW,HEL,&MGME2);
	  }
	  // h->qqbar'q'gqbar
	  else if (id[4] == 21) {
	    coupl_mass_.bmass = m[1];
	    coupl_mass_.cmass = m[2];
	    sh2bcxcgbx_(&colorDepth,PNEW,HEL,&MGME2);
	  }
        }
     
        if (verbose >= 8) {
            printOut("VinciaMGInterface::ME2","Getting M5 with colDepth = "
                        +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
        }
            
      }
      
      // h->6
      else if (nOut == 6) {
	
        // Convert momenta to MadGraph format
        double PNEW[7][4];
        for (int i=0; i<int(nMom); ++i) Vec4toMG(p[i],PNEW[i]);
	
        // HELCOMB = 1 + (pol[1] + 1)*16 + ... + (pol[6] + 1)/2
        // 0) Default assumption : fully unpolarized h decay
        // 1) Fully polarized h decay
        if (pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
	    && pol[4] != 9 && pol[5] != 9 && pol[6] != 9) {
          HEL[0] = (int)(1 + (pol[6] + 1)/2 + (pol[5] + 1) 
                         + 2*(pol[4] + 1) + 4*(pol[3] + 1) + 8*(pol[2] + 1) 
                         + 16*(pol[1] + 1));
          HEL[1] = -1;
          HEL[2] = -1;
        }
	
        //NOTE: h>gggggg matrix elements are not included
        
        //h>qggggqbar
        if (cSum == 8 && (id[1]+id[6]==0)) {
          coupl_mass_.bmass = m[1];
          sh2bggggbx_(&colorDepth,PNEW,HEL,&MGME2);
        }
	
        // h->q qbar' q' g g qbar
        else if (cSum == 4 && (id[2]+id[3]==0)) {
	  // the matrix element has no h-u-type coupling
	  // set b-mass to that of primary quark, c to secondary
	  coupl_mass_.bmass = m[1];
	  coupl_mass_.cmass = m[2];
	  sh2bcxcggbx_(&colorDepth,PNEW,HEL,&MGME2);
        }
        // h->g q qbar' q' g qbar
        else if (cSum == 4 && (id[3]+id[4]==0)) {
	  // the matrix element has no h-u-type coupling
	  // set b-mass to that of primary quark, c to secondary
	  coupl_mass_.bmass = m[1];
	  coupl_mass_.cmass = m[3];
	  sh2bgcxcgbx_(&colorDepth,PNEW,HEL,&MGME2);
        }
        // h->q g g qbar' q' qbar
        else if (cSum == 4 && (id[4]+id[5]==0)) {
	  // the matrix element has no h-u-type coupling
	  // set b-mass to that of primary quark, c to secondary
	  coupl_mass_.bmass = m[1];
	  coupl_mass_.cmass = m[4];
	  sh2bggcxcbx_(&colorDepth,PNEW,HEL,&MGME2);
        }
        // h->q q' qbar' q'' qbar'' qbar
        // the MG interface does not allow at present to make a
        // light quark massive, therefore one of the secondary quarks
        // involved has to be massless
        else if (cSum == 0 && id[2]+id[3]==0 && id[4]+id[5]==0) {
	  // the matrix element has no h-u-type coupling
	  if (m[5] == 0.0) {
	    coupl_mass_.bmass = m[1];
	    coupl_mass_.cmass = m[3];
	    sh2bcxcuxubx_(&colorDepth,PNEW,HEL,&MGME2);
	  }
	  else if (m[3] == 0.0) {
	    coupl_mass_.bmass = m[1];
	    coupl_mass_.cmass = m[5];
	    sh2buxucxcbx_(&colorDepth,PNEW,HEL,&MGME2);
	  }
	  // case of both secondary quarks massive not implemented
        }
          
        if (verbose >= 8) {
          printOut("VinciaMGInterface::ME2","Getting M6 with colDepth = "
                   +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
        }
      }
        
    }

    if (isnan(MGME2) && verbose >= 1) {
      printOut("VinciaMGInterface::ME2","Error! ME2 = nan. Printing momenta:");
      for (int i=0; i<int(p.size()); ++i) {
	cout<<"   p"<<i<<p[i];
      }
    }

    // Return |ME|^2 value (or -1 if no ME found)
    if (abs(MGME2+1) < TINY) return -1.;
    else return MGME2/max(1,nPolIn);

  }

}
